<?php
/**
 * @file
 * Implementation of cache.inc with memcache logic included.
 */

/**
 * Memcache class extended to support Cache Heuristic.
 */
class MemCacheDrupalDeferred extends MemCacheDrupal implements DrupalCacheDeferredInterface {
  /**
   * Serialize the data.
   *
   * @var boolean
   */
  protected $serialize = TRUE;

  /**
   * Constructor.
   */
  public function __construct($bin) {
    $this->serialize = cache_heuristic_variable_get('memcache_deferred', $bin, 'serialize', TRUE);
    parent::__construct($bin);
  }

  /**
   * Implements DrupalCacheInterface::get().
   */
  public function get($cid) {
    $cache = dmemcache_get($cid, $this->bin, $this->memcache);
    if ($this->valid($cid, $cache)) {
      $cache = $this->prepareItem($cache);
      return $cache;
    }
    return FALSE;
  }

  /**
   * Implements DrupalCacheInterface::getMultiple().
   */
  public function getMultiple(&$cids) {
    $results = dmemcache_get_multi($cids, $this->bin, $this->memcache);
    foreach ($results as $cid => &$result) {
      if (!$this->valid($cid, $result)) {
        // This object has expired, so don't return it.
        unset($results[$cid]);
      }
      else {
        $result = $this->prepareItem($result);
      }
    }
    // Remove items from the referenced $cids array that we are returning,
    // per the comment in cache_get_multiple() in includes/cache.inc.
    $cids = array_diff($cids, array_keys($results));
    return $results;
  }

  /**
   * Implements DrupalCacheInterface::set().
   */
  public function set($cid, $data, $expire = CACHE_PERMANENT) {
    $created = time();

    // Create new cache object.
    $cache = new stdClass();
    $cache->cid = $cid;
    $cache_data = is_object($data) ? clone $data : $data;
    $cache->data = $this->serialize ? serialize($cache_data) : $cache_data;
    $cache->serialized = $this->serialize;
    $cache->created = $created;
    // Record the previous number of wildcard flushes affecting our cid.
    $cache->flushes = $this->wildcard_flushes($cid);
    if ($expire == CACHE_TEMPORARY) {
      // Convert CACHE_TEMPORARY (-1) into something that will live in memcache
      // until the next flush.
      $cache->expire = REQUEST_TIME + 2591999;
    }
    // Expire time is in seconds if less than 30 days, otherwise is a timestamp.
    elseif ($expire != CACHE_PERMANENT && $expire < 2592000) {
      // Expire is expressed in seconds, convert to the proper future timestamp
      // as expected in dmemcache_get().
      $cache->expire = REQUEST_TIME + $expire;
    }
    else {
      $cache->expire = $expire;
    }

    // Manually track the expire time in $cache->expire.  When the object
    // expires, if stampede protection is enabled, it may be served while one
    // process rebuilds it. The ttl sent to memcache is set to the expire twice
    // as long into the future, this allows old items to be expired by memcache
    // rather than evicted along with a sufficient period for stampede
    // protection to continue to work.
    if ($cache->expire == CACHE_PERMANENT) {
      $memcache_expire = $cache->expire;
    }
    else {
      $memcache_expire = $cache->expire + (($cache->expire - REQUEST_TIME) * 2);
    }
    dmemcache_set($cid, $cache, $memcache_expire, $this->bin, $this->memcache);
  }

  /**
   * Unserialize the data if applicable.
   */
  public function prepareItem($item) {
    // If the data is permanent or not subject to a minimum cache lifetime,
    // unserialize and return the cached data.
    if (!empty($item->serialized)) {
      $item->data = unserialize($item->data);
    }

    return $item;
  }

  /**
   * Implements DrupalCacheDeferredInterface::getMultipleDeferred().
   */
  public function getMultipleDeferred($cids) {
    $results = dmemcache_get_multi($cids, $this->bin, $this->memcache);
    return $results;
  }

  /**
   * Implements DrupalCacheDeferredInterface::prepareDeferredItem().
   */
  public function prepareDeferredItem($cid, $item) {
    return $this->prepareItem($item);
  }

  /**
   * Implements DrupalCacheDeferredInterface::fetchDeferredItems().
   */
  public function fetchDeferredItems($cids, $results) {
    $results = $results ? $results : array();
    foreach ($results as $cid => &$result) {
      if (!$this->valid($cid, $result)) {
        // This object has expired, so don't return it.
        unset($results[$cid]);
      }
    }
    return $results;
  }

}
